#include "U8glib.h"
#include <avr/interrupt.h>
#include <avr/io.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>

/* ----------------------------- JOYSTICK CONFIG --------------------------- */
#define VRX_PIN         A0        // Arduino pin connected to VRX pin
#define VRY_PIN         A1        // Arduino pin connected to VRY pin
#define BUTTON_PIN      PD2       // Arduino pin connected to SW pin

#define LEFT_THRESHOLD  200
#define RIGHT_THRESHOLD 823
#define UP_THRESHOLD    200
#define DOWN_THRESHOLD  823

#define COMMAND_NO      0x00
#define COMMAND_LEFT    0x01
#define COMMAND_RIGHT   0x02
#define COMMAND_UP      0x04
#define COMMAND_DOWN    0x08

int xValue = 0 ;                  // Value of the X axis
int yValue = 0 ;                  // Value of the Y axis
int command = COMMAND_NO;         // Calculated using xValue & yValue  
int lastCommand = 0;              // Used for joystick debouncing

unsigned long btnDelay = 50;      // Debounce delay in milliseconds
unsigned long btnLastTime = 0;   // Time stamp of the last button state change


/* ------------------------------ BUZZER CONFIG ---------------------------- */
#define BUZZER_PIN      9

/* -------------------------------- UI CONFIG ------------------------------ */
#define NOT_STARTED     0
#define IN_PROGRESS     1
#define VICTORY         -1
#define GAME_OVER       -2

U8GLIB_ST7920_128X64_4X u8g(10);  // LCD declaration

bool introAnimation = false;      // Checks if the initial page was loaded

static char screenValues[12][5] = {"", "   2", "   4", "   8", 
                                   "  16", "  32", "  64", " 128",
                                   " 256", " 512", "1024", "2048"};

volatile unsigned long elapsedTime = 0;
static int8_t gameState;


/* ---------------------------- 2048 GAME CONFIG --------------------------- */
#define EMPTY_SLOT    0
#define LEFT          1
#define RIGHT         2
#define UP            3
#define DOWN          4

uint8_t SIZE = 4;
#define MAX_SIZE      4
uint8_t grid[MAX_SIZE][MAX_SIZE];

/* ------------------------------------------------------------------------- */
/* --------------------------- CONFIGURATION DONE -------------------------- */


/* ----------------------- 2048 LOGIC FLOW FUNCTIONS ----------------------- */

static uint8_t tileIsMoving;
static uint8_t moveType;
static uint8_t move_i, move_j;

void gameInit(void)
{
  memset(grid, 0, MAX_SIZE * MAX_SIZE * sizeof(uint8_t));
  createTile();
  createTile();
}

void createTile()
{
  uint8_t value, i, j;
  
  uint8_t r = rand() % 256;
  if (r > 230)
    value = 2;
  else
    value = 1;
      
  do {
    r = rand() % (SIZE * SIZE);
    i = r / SIZE;
    j = r % SIZE;
  } while (grid[i][j] != EMPTY_SLOT);
  
  grid[i][j] = value;
}

void moveLeft(uint8_t i)
{
  uint8_t c = 0;
  uint8_t j = 0;
  uint8_t u = 0;
    
  for (j = 0; j < SIZE; ++j)
  {
    c = j;
    if (grid[i][j] == EMPTY_SLOT)
    {
      while (grid[i][++c] == EMPTY_SLOT && c < SIZE);
      if (c >= SIZE)
        break;
      
      grid[i][j] = grid[i][c];
      grid[i][c] = EMPTY_SLOT;
    }
    
    if (j > 0 && u == 0 && grid[i][j - 1] == grid[i][j])
    {
      u = 1;
      ++grid[i][j - 1];
      grid[i][j] = EMPTY_SLOT;
      --j;
    }
    else
      u = 0;
  }
}

void mirrorOX(uint8_t m[MAX_SIZE][MAX_SIZE])
{
  for (unsigned char i = 0; i < SIZE; ++i) {
    unsigned char t = m[i][0];
    m[i][0] = m[i][SIZE - 1];
    m[i][SIZE - 1] = t;
    t = m[i][1];
    m[i][1] = m[i][SIZE - 2];
    m[i][SIZE - 2] = t;
  }
}

void rotateMatrix(uint8_t m[MAX_SIZE][MAX_SIZE], uint8_t direction)
{
  uint8_t new_m[MAX_SIZE][MAX_SIZE];
  for (unsigned char i = 0; i < MAX_SIZE; ++i) {
    memset(new_m[i], 0, MAX_SIZE);
  }
  
  if (direction == LEFT) {
    for (unsigned char i = 0; i < SIZE; ++i)
      for (unsigned char j = 0; j < SIZE; ++j)
        new_m[i][j] = m[j][SIZE - i - 1];
  } else {
    for (unsigned char i = 0; i < SIZE; ++i)
      for (unsigned char j = 0; j < SIZE; ++j)
        new_m[i][j] = m[SIZE - j - 1][i];
  }
  
  memcpy(m, new_m, MAX_SIZE * MAX_SIZE * sizeof(uint8_t));
}

void beginMoveLCD(unsigned char type)
{
  tileIsMoving = 1;
  moveType = type;
  if (type == LEFT) {
    move_i = 0;
    move_j = -1;
  } else if (type == RIGHT) {
    move_i = 0;
    move_j = 1;
  } else if (type == UP) {
    move_i = -1;
    move_j = 0;
  } else if (type == DOWN) {
    move_i = 1;
    move_j = 0;
  }
}

void move(uint8_t direction)
{
  if (direction == RIGHT) {
    mirrorOX(grid);
  } else if (direction == UP) {
    rotateMatrix(grid, LEFT);
  } else if (direction == DOWN) {
    rotateMatrix(grid, RIGHT);
  }
  
  for (unsigned char i = 0; i < SIZE; ++i)
    moveLeft(i);

  if (direction == RIGHT) {
    mirrorOX(grid);
  } else if (direction == UP) {
    rotateMatrix(grid, RIGHT);
  } else if (direction == DOWN) {
    rotateMatrix(grid, LEFT);
  }
}

uint8_t moveIsCompleted(uint8_t type)
{
  uint8_t temp[MAX_SIZE][MAX_SIZE];
  memcpy(temp, grid, MAX_SIZE * MAX_SIZE * sizeof(uint8_t));
  
  move(type);
  
  uint8_t res = memcmp(temp, grid, MAX_SIZE * MAX_SIZE * sizeof(uint8_t));
  memcpy(grid, temp, MAX_SIZE * MAX_SIZE * sizeof(uint8_t));
  
  return res;
}

uint8_t evaluateGame(void)
{
  if (!moveIsCompleted(LEFT) &&
      !moveIsCompleted(RIGHT) &&
      !moveIsCompleted(UP) &&
      !moveIsCompleted(DOWN))
    return GAME_OVER;
  
  for (unsigned char i = 0; i < SIZE; ++i)
    for (unsigned char j = 0; j < SIZE; ++j)
      if (SIZE == 4 && grid[i][j] == 11 ||
          SIZE == 3 && grid[i][j] == 7  ||
          SIZE == 2 && grid[i][j] == 4)
        return VICTORY;
              
  return IN_PROGRESS;
}

/* ------------------------------------------------------------------------- */


/* ------------------------------ UI Functions ----------------------------- */

/* Makes the buzzer sound for each move */
void buzzerSound() {
  unsigned long startTime = millis();
  digitalWrite(BUZZER_PIN, LOW);
  while (millis() - startTime < 15) ;
  digitalWrite(BUZZER_PIN, HIGH);
}

/* Converts the elapsedTime(seconds) to "mm:ss" and prints it */
void printElapsedTime() {
  unsigned long minutes = elapsedTime / 60;
  unsigned long remainingSeconds = elapsedTime % 60;

  const char formattedTime[10];
  sprintf(formattedTime, "%lu:%02lu", minutes, remainingSeconds);

  u8g.drawStr(30, 62, "TIME:");
  u8g.drawStr(72, 62, formattedTime);
}

/* Clears the LCD display */
void clearLCD() {
    u8g.firstPage();  
    do {
    } while( u8g.nextPage() );
}

/* Draws the game's grid */
void drawGrid() {
  /* Draw grid */
  u8g.setDefaultForegroundColor();
  u8g.drawFrame(0,0,128,64);

  u8g.drawLine(31, 0, 31, 63);
  u8g.drawLine(32, 0, 32, 63);
  u8g.drawLine(63, 0, 63, 63);
  u8g.drawLine(64, 0, 64, 63);
  u8g.drawLine(95, 0, 95, 63);
  u8g.drawLine(96, 0, 96, 63);

  u8g.drawLine(0, 15, 127, 15);
  u8g.drawLine(0, 16, 127, 16);
  u8g.drawLine(0, 31, 127, 31);
  u8g.drawLine(0, 32, 127, 32);
  u8g.drawLine(0, 47, 127, 47);
  u8g.drawLine(0, 48, 127, 48);
}

/* Renders the current frame */
void render(void) {
  if (gameState == NOT_STARTED) {
    /* 2048 intro page */
    if (!introAnimation) {
      u8g.firstPage();
      do {
        u8g.drawFrame(0,0,128,64);
        u8g.setFont(u8g_font_osr35n);
        u8g.drawStr(7, 48, "2048");
      } while( u8g.nextPage() );
      delay(3500);
      clearLCD();

      introAnimation = true;
    }

    /* PICK THE DESIRED GAME */ 
    u8g.setDefaultForegroundColor();
    u8g.setFont(u8g_font_helvB10);
    u8g.drawStr(8, 25, "Pick your game!");
    u8g.setFont(u8g_font_helvR08);
    if (SIZE == 4) {
      u8g.setDefaultForegroundColor();
      u8g.drawBox(17, 40, 30, 14);
      u8g.setDefaultBackgroundColor();
      u8g.drawStr(24, 51, "4x4");

      u8g.setDefaultForegroundColor();
      u8g.drawStr(56, 51, "3x3");
      u8g.drawStr(88, 51, "2x2");
    } else if (SIZE == 3) {
      u8g.setDefaultForegroundColor();
      u8g.drawBox(49, 40, 30, 14);
      u8g.setDefaultBackgroundColor();
      u8g.drawStr(56, 51, "3x3");

      u8g.setDefaultForegroundColor();
      u8g.drawStr(24, 51, "4x4");
      u8g.drawStr(88, 51, "2x2");
    } else if (SIZE == 2) {
      u8g.setDefaultForegroundColor();
      u8g.drawBox(81, 40, 30, 14);
      u8g.setDefaultBackgroundColor();
      u8g.drawStr(88, 51, "2x2");

      u8g.setDefaultForegroundColor();
      u8g.drawStr(24, 51, "4x4");
      u8g.drawStr(56, 51, "3x3");
    }
  } else if (gameState == VICTORY) {
    u8g.setDefaultForegroundColor();
    u8g.setFont(u8g_font_helvB10);
    u8g.drawStr(30, 14, "VICTORY!");
    u8g.drawLine(0, 16, 128, 16);

    u8g.setFont(u8g_font_courR08);
    u8g.drawStr(17, 30, "Press the button");
    u8g.drawStr(35, 38, "to restart");
    u8g.drawLine(0, 48, 128, 48);

    u8g.setFont(u8g_font_profont11);
    printElapsedTime();
  } else if (gameState == GAME_OVER) {
    u8g.setDefaultForegroundColor();
    u8g.setFont(u8g_font_helvB10);
    u8g.drawStr(18, 14, "GAME OVER!");
    u8g.drawLine(0, 16, 128, 16);

    u8g.setFont(u8g_font_courR08);
    u8g.drawStr(17, 30, "Press the button");
    u8g.drawStr(35, 38, "to restart");
    u8g.drawLine(0, 48, 128, 48);

    u8g.setFont(u8g_font_profont11);
    printElapsedTime();
  } else {
    /* STILL IN GAME */
    u8g.setFont(u8g_font_helvR08);

    drawGrid();

    /* `cut` unused slots */
    for (unsigned char i = SIZE; i < MAX_SIZE; ++i) {
      for (unsigned char j = 0; j < MAX_SIZE; ++j) {
        u8g.drawLine(j * 32 + 1, i * 16 + 1, j * 32 + 31, i * 16 + 15);
        u8g.drawLine(i * 32 + 1, j * 16 + 1, i * 32 + 31, j * 16 + 15);
      }
    }

    /* render the numbers */
    for (unsigned char i = 0; i < SIZE; ++i)
      for (unsigned char j = 0; j < SIZE; ++j)
        if (grid[i][j] > 0) {      
          u8g.drawStr(j * 32 + 4 + tileIsMoving * move_j * 10,
                      i * 16 + 12 + tileIsMoving * move_i * 5,
                      screenValues[grid[i][j]]);
        }
  }
}


/* Converts the analogic joystick's input to a game command */
int getCurrentCommand() {
  int command = COMMAND_NO;

  /* Get the analog values of X and Y axis */
  xValue = analogRead(VRX_PIN);
  yValue = analogRead(VRY_PIN);

  /* Convert the analog values to according commands */
  if (xValue < LEFT_THRESHOLD) {
    command = command | COMMAND_LEFT;
  } else if (xValue > RIGHT_THRESHOLD) {
    command = command | COMMAND_RIGHT;
  }

  if (yValue < UP_THRESHOLD) {
    command = command | COMMAND_UP;
  } else if (yValue > DOWN_THRESHOLD) {
    command = command | COMMAND_DOWN;
  }

  return command;
}

/* Executes the command received from input */
void executeCommand(int command) {
  if (!tileIsMoving) {
    /* Block moves between transitions */
    if (!(lastCommand & command)) {
      /* Block multiple moves on a single input */
      if (command & COMMAND_UP) {
        if (moveIsCompleted(UP)) {
          beginMoveLCD(UP);
          buzzerSound();
        }
      } else if (command & COMMAND_LEFT) {
        if (moveIsCompleted(LEFT)) {
          beginMoveLCD(LEFT);
          buzzerSound();
        }
      } else if (command & COMMAND_DOWN) {
        if (moveIsCompleted(DOWN)) {
          beginMoveLCD(DOWN);
          buzzerSound();
        }
      } else if (command & COMMAND_RIGHT) {
        if (moveIsCompleted(RIGHT)) {
          beginMoveLCD(RIGHT);
          buzzerSound();
        }
      }

      lastCommand = command;
    }
  }
}


/* Button interruption routine used to start/restart the game */
ISR(INT0_vect) {
  if (millis() - btnLastTime >= btnDelay) {
    btnLastTime = millis();  // Update the last button state change time

    if (gameState != NOT_STARTED) {
      gameState = NOT_STARTED;
    } else {
      gameState = IN_PROGRESS;
      elapsedTime = 0;
      srand((unsigned)TCNT1);
      gameInit();
    }
  }
}

/* Timer1 overflow interrupt service routine
   increments elapsed time once every second */
ISR(TIMER1_OVF_vect) {
  if (gameState == IN_PROGRESS)
    elapsedTime++;
}


void setup() {
  /* BUZZER SETUP */
  pinMode (BUZZER_PIN, OUTPUT);
  digitalWrite(BUZZER_PIN, HIGH);

  /* IO SETUP - JOYSTICK BUTTON */
  DDRD &= ~(1 << BUTTON_PIN);     // Set the button pin as input
  PORTD |= (1 << BUTTON_PIN);     // Enable internal pull-up resistor for the button pin

  EICRA |= (1 << ISC01);          // Set interrupt on falling edge
  EIMSK |= (1 << INT0);           // Enable external interrupt INT0

  /* TIMER1 SETUP */
  TCCR1A = 0;                     // Set Timer1 to normal mode
  TCCR1B = (1 << CS12);           // Set prescaler to 256
  TIMSK1 = (1 << TOIE1);          // Enable Timer1 overflow interrupt
  TCNT1 = 0;                      // gameInit Timer1 count to 0
  gameState = NOT_STARTED;

  sei();                          // Enable global interrupts

  // debug
  // Serial.begin(9600);
}

void debugFunc() {
  if (command != COMMAND_NO) {
    for (int i = 0; i < MAX_SIZE; i++) {
      for (int j = 0; j < MAX_SIZE; j++) {
        Serial.print(grid[i][j]);
        Serial.print(" ");
      }
      Serial.println();
    }
    Serial.println();
  }
}

void loop() {
  /* Read joystick's current state - toggle the buzzer if the case */
  command = getCurrentCommand();

  /* START MENU - PICK THE MODE */
  if (gameState == NOT_STARTED) {
    if (command & COMMAND_RIGHT && !(lastCommand & command)) {
      if (SIZE > 2) {
        SIZE--;
      }
    } else if (command & COMMAND_LEFT && !(lastCommand & command)) {
      if (SIZE < 4) {
        SIZE++;
      }
    }
  }
  
  /* Execute the joystick's input command */
  executeCommand(command);

  // debugFunc();

  /* Check if the game is finished and set proper value for gameState variable */
  if (gameState == IN_PROGRESS) {
    gameState = evaluateGame();
  }
  
  /* Render current frame on LCD */
  u8g.firstPage();
  do {
    render();
  } while (u8g.nextPage());

  /* Process the current move and insert a new tile */
  if (tileIsMoving) {
    tileIsMoving = 0;
    move(moveType);
    createTile();
  }
}
